package oeg.merero;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.recommender101.data.DataModel;
import org.recommender101.data.DefaultDataLoader;
import org.recommender101.data.Rating;
import org.recommender101.eval.impl.Recommender101Impl;
import org.recommender101.recommender.AbstractRecommender;
import org.recommender101.recommender.extensions.funksvd.FunkSVDRecommender;
import org.recommender101.tools.Utilities101;

import oeg.merero.recommenders.BestRecommender;
import oeg.merero.recommenders.FlipRecommender;
import oeg.merero.recommenders.MaxMSERecommender;
import oeg.merero.recommenders.evaluator.MEREROEvaluator;


public class MinimalTest {

	public static void main(String[] args) {
		try {
			
			Recommender101Impl.TOP_N = 100000;
			int minRank = 1;
			int maxRank = 5;
			/*
			 * ***********************************
			 * Leemos nuestros datos
			 * ***********************************
			 */
			DefaultDataLoader dataLoader = new DefaultDataLoader();
			DataModel allData = new DataModel();
			
			allData.setMinRatingValue(minRank);
			allData.setMaxRatingValue(maxRank);

			dataLoader.setFilename("data/movielens/MovieLens100kRatings.txt");
			dataLoader.setSampleNUsers("100000");
			dataLoader.loadData(allData);

			/*
			 * ****************************************
			 * Vamos a probar sobre el usuario 13 es 
			 * el que tiene mas calificaciones emitidas
			 * *****************************************
			 * */
			int testUser = 13;
			Set<Rating> uRatings = allData.getRatingsOfUser(testUser);
			System.out.println("EL usuairo tiene n :" + uRatings.size());

			
			
			/*
			 * *********************************************
			 * Creamos el conjunto de entrenamuento 
			 * 
			 * Todos - los del items de usuario 13 que 
			 * queremos estimar
			 * *********************************************
			 */
			DataModel testDataModel = new DataModel();
			DataModel trainingDataModel = new DataModel(allData);

			trainingDataModel.setMinRatingValue(minRank);
			trainingDataModel.setMaxRatingValue(maxRank);

			
			
				
			// Tomamos 10 items al azar para las pruebas y 
			// los quitamos del conjunto de entrenamiento
			int[] itemsToEval = {561, 312, 866, 755, 79, 38, 770, 739, 272, 525};
			//ArrayList<Rating> arrayRate = new ArrayList<Rating>(uRatings);
			//Collections.shuffle(arrayRate);
			for (int i = 0; i < itemsToEval.length; i++) {
				//Rating rate = arrayRate.get(i);
				int idItem = itemsToEval[i];
				float rateV = trainingDataModel.getRating(testUser, idItem);
				Rating rate = new Rating(testUser, idItem, rateV);
				
				testDataModel.addRating(rate);
				trainingDataModel.removeRating(rate);
			}
			
			
			/*
			 * ***********************************
			 * Instanciamos nuestros SR de prueba
			 * ***********************************
			 */
			FlipRecommender		flip	= new FlipRecommender();
			MaxMSERecommender	maxMSE	= new MaxMSERecommender();
			BestRecommender		best	= new BestRecommender();
			FunkSVDRecommender	funkSVD	= new FunkSVDRecommender();
			funkSVD.setNumFeatures("50");
			funkSVD.setInitialSteps("50");
			

			
			Hashtable<String, Map> expResults = new Hashtable<String, Map>();
			AbstractRecommender[] recommenderList = new AbstractRecommender[] { best, flip, maxMSE, funkSVD };
			String txtOutTraining = ";*************TIEMPOR ENTRENAMIENTO*******************;\n";
			String txtOutRanking = ";*************EVALUACION*******************;\n";
			
			
			System.out.println("Entrenando modelos ...");
			String origName = best.getClass().getSimpleName();
			for (int i = 0; i < recommenderList.length; i++) {
				AbstractRecommender recommender = recommenderList[i];
				String rsName = recommender.getClass().getSimpleName();
				/*
				 * ***********************************
				 * Entrenamos el SR
				 * ***********************************
				 */
					long time = (new Date()).getTime();
				recommender.setDataModel(trainingDataModel);
				recommender.setExtraInformation("OriginalTestData", allData);
				recommender.setExtraInformation("TestData", testDataModel);
				recommender.init();
					time = (new Date()).getTime()- time;
				
				txtOutTraining += rsName+"\t: "+time+" ms\n";
				
				// Donde guardamos las estimaciones
				Map<Integer, Float> estimedRatings = new HashMap<Integer, Float>();
				// Los items que vamos a usar para le prueba
				Iterator<Integer> testItems = testDataModel.getItems().iterator();
				while(testItems.hasNext()) {
					int tmpItem = testItems.next();
					float r = recommender.predictRating(testUser, tmpItem);
					estimedRatings.put(tmpItem, r);
				}

				estimedRatings = Utilities101.sortByKeyAscending(estimedRatings);
				estimedRatings = Utilities101.sortByValueDescending(estimedRatings);
				
				expResults.put(rsName, estimedRatings); 
				
				@SuppressWarnings("unchecked")
				Map<Integer, Float> realValues = expResults.get(origName);
				
				
				List<Integer> ranking = new ArrayList<Integer>(estimedRatings.keySet());
				List<Integer> rankOrig = new ArrayList<Integer>(realValues.keySet());
				
				
				float mseDiff = MinimalTest.calculaMSE(realValues, estimedRatings);
				float mereroDiff = MEREROEvaluator.calcDistance(ranking, rankOrig);
				
				
				Map<Integer, Integer> vectPredicted = Utilities101.sortByKeyAscending(MEREROEvaluator.convertRankingToPositionVector(ranking));
				Map<Integer, Integer> vectOriginal = Utilities101.sortByKeyAscending(MEREROEvaluator.convertRankingToPositionVector(rankOrig));
				
				
				List<Integer> tmpKeys = new ArrayList<Integer>(vectOriginal.keySet());
				String sumaMh = "";
				String sumaMh2 = "";
				for (Integer itemID : tmpKeys) {
					sumaMh += "|"+vectPredicted.get(itemID)+"-"+vectOriginal.get(itemID)+"|+";
					sumaMh2 += Math.abs(vectPredicted.get(itemID)-vectOriginal.get(itemID))+" + ";
				}
				
				
				
				txtOutRanking += "\n"+rsName+ " * [merero:"+mereroDiff+", mse:"+mseDiff+"]:\n" +
						"\tEstimaciones (<idItem>=<valorEstimado>):\n"+
						"\t\t"         +estimedRatings+"\n"+
						"\tRankings:\n"+
						"\t\tOriginal: "+rankOrig+"\n"+
						"\t\tEstimado: "+ranking+"\n"+
						" \tVectores: \n"+
						"\t\tOriginal   : "+vectOriginal+"\n"+
						"\t\tEstimado   : "+vectPredicted+"\n"+
						"\t\tDiferencias: "+sumaMh+"\n"+
						"\t\t           : "+sumaMh2+"\n"+
						"\t\t           : "+mereroDiff+"\n"+
						"";
			}

			
			/*
			 * *************************************************
			 * Imprimiendo resultados
			 * *************************************************
			 */
			System.out.println(txtOutTraining);
			System.out.println(txtOutRanking);
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}
	
	
	/**
	 * Funcion auxiliar para calcular el error cuadratico medio
	 * @param original Algo como [it1=3, it2=3]
	 * @param predicted Algo como [it1=3.2, it2=4.5]
	 * @return El error cuadratico medio entre las dos listas o NaN si no todos los items de original estan en predicted o si original es vacio.
	 */
	public static float calculaMSE(Map<Integer, Float> original, Map<Integer, Float> predicted) {
		float mse = 0 ;
		if(original.size() <= 0 || predicted.size() <= 0) {
			return Float.NaN;
		}
			
		Iterator<Integer> items = original.keySet().iterator();
		while(items.hasNext()) {
			Integer idxItem = items.next();
			Float origV =  original.get(idxItem);
			Float predV = predicted.get(idxItem); 
			if(predV == null || predV < 0) { //No se encontro en el otro conjunto
				return Float.NaN;	
			}
			
			float error = origV - predV;
			mse += error*error;
		}
		return mse/original.size();
	}

}
