/**
 * 
 */
package oeg.merero.recommenders;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.recommender101.data.DataModel;
import org.recommender101.data.Rating;
import org.recommender101.recommender.AbstractRecommender;
import org.recommender101.tools.Utilities101;

/**
 * @author oescamil
 *
 */
abstract public class AbstractTestRecommender extends AbstractRecommender {	
	static final long serialVersionUID = -3073503454386601618L;
	/**
	 * The mean of the 
	 */
	protected float r_med = 0;
	protected float r_max = 0;
	protected float r_min = 0;
	
	/**
	 * 
	 * @return
	 */
	protected DataModel getOriginalDataModel(){
		return (DataModel) this.getExtraInformation("OriginalTestData");
	} 
	
	
	
	/* (non-Javadoc)
	 * @see org.recommender101.recommender.AbstractRecommender#recommendItems(int)
	 */
	@Override
	public List<Integer> recommendItems(int user) {
		List<Integer> result = this.recommendItemsByRatingPrediction(user);
		return result;
	}

	/**
	 * A general method for ranking items according to their rating prediction.
	 * The method should be overwritten in case the recommender cannot make 
	 * rating predictions or when a better heuristic is needed, which for 
	 * example takes the popularity of the recommendations into account.
	 * @param user the user for which a recommendation is sought
	 * @return the ranked list of items
	 */
	@Override
	public List<Integer> recommendItemsByRatingPrediction(int user) {
		List<Integer> result = new ArrayList<Integer>();

		// If there are no ratings for the user in the test set,
		// there is no point of making a recommendation.
		Map<Integer, Set<Rating>> ratings = getDataModel().getRatingsPerUser();
		// If we have no ratings...
		if (ratings == null || ratings.size() == 0) {
			return Collections.emptyList();
		}

		
		DataModel origModel = getOriginalDataModel();
		// Calculate rating predictions for all items we know
		Map<Integer, Float> predictions = new HashMap<Integer, Float>();
		//byte rating = -1;
		float pred = Float.NaN;
		// Go through all the items
		for (Integer item : origModel.getItems()) {
			
			float rating = getDataModel().getRating(user, item);
			// We will not recommend items repeatedly here-
			if (rating == -1 || Float.isNaN(rating)) {
				// make a prediction and remember it in case the recommender
				// could make one
				pred = predictRating(user, item);
				if (!Float.isNaN(pred)) {
					predictions.put(item, pred);
				}
			}
		}
		
		predictions = filterElementsByRelevanceThreshold(predictions, user);
		predictions = Utilities101.sortByKeyAscending(predictions);
		predictions = Utilities101.sortByValueDescending(predictions);
		
		for (Integer item : predictions.keySet()) {
			result.add(item);
		}
		return result;
	}
	
	/**
	 * Initialize the recommender. Calculates the mean from the ratings 
	 * @see org.recommender101.recommender.AbstractRecommender#init()
	 */
	@Override
	public void init() throws Exception {
		DataModel dataModel = this.getDataModel();
		
		r_max = dataModel.getMaxRatingValue();
		r_min = dataModel.getMinRatingValue();
		r_med = (r_max + r_min)/2.0f;
		
		
	}	
}
