/**
 * 
 */
package oeg.merero.recommenders;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.recommender101.data.DataModel;
import org.recommender101.data.Rating;
import org.recommender101.eval.impl.Recommender101Impl;
import org.recommender101.tools.Utilities101;

/**
 * @author oescamil
 *
 */
public class FlipRecommender extends AbstractTestRecommender {	
	static final long serialVersionUID = -3073503454386601618L;
	private HashMap<Integer, List<Integer>> orinalRankings;
	private int itemCount = 0;
	
	/* (non-Javadoc)
	 * @see org.recommender101.recommender.AbstractRecommender#predictRating(int, int)
	 */
	@Override
	public void init() throws Exception {
		super.init();
		orinalRankings = new HashMap<Integer, List<Integer>>();
		DataModel origModel = this.getOriginalDataModel();
		this.itemCount = origModel.getItems().size();
	}




	@Override
	public float predictRating(int user, int item) {
		DataModel origModel = this.getOriginalDataModel();
		
		if(orinalRankings.get(user) == null) {
			orinalRankings.put(user,createOriginalRankingFor(user));
		}
		
		List<Integer> origRank = orinalRankings.get(user);
		if(origRank.indexOf(item) < 0 ) {
			return Float.NaN;
		}
		
		float r_ui = origModel.getRating(user, item);
		if(r_ui <= -1){
			return Float.NaN;
		}
		
		float newRui= r_min+(r_max-r_ui)+((float)item/(float)itemCount)*0.0001f;
		return newRui;
	}

	/**
	 * Crea el ranking original de items del usuario, solo toma en cuenta 
	 * las primeras Recommender101Impl.TOP_N posiciones del ranking.
	 * @param user El usuario de cual queremos el ranking original
	 * @return
	 */
	private List<Integer> createOriginalRankingFor(int user) {
		int topN = Recommender101Impl.TOP_N;
		DataModel testModel = (DataModel) this.getExtraInformation("TestData");
		DataModel trainingM = this.getDataModel();
		Set<Rating> userRatings = testModel.getRatingsOfUser(user);
		
		// Get rating for all items we know
		Map<Integer, Float> originalRatings = new HashMap<Integer, Float>();
		for(Rating r : userRatings){
			float trainingR = trainingM.getRating(r.user, r.item); 
			if(trainingR == -1) // Just for items which recommender can recommend
				originalRatings.put(r.item, r.rating);
		}
		originalRatings = Utilities101.sortByKeyAscending(originalRatings);
		originalRatings = Utilities101.sortByValueDescending(originalRatings);
		List<Integer> originalRanking = new ArrayList<Integer>();
		
		topN = Math.min(originalRatings.size(), topN);
		int counter = 0 ; 
		for (Integer item : originalRatings.keySet()) {
			originalRanking.add(item);
			counter++;
			if(counter >= topN)
				break;
		}
		List<Integer> res = originalRanking.subList(0, topN);
		originalRanking = null;
		//System.gc();
		return res;
	}	
}
