/**
 * 
 */
package oeg.merero.recommenders.evaluator;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.management.ImmutableDescriptor;
import javax.print.attribute.IntegerSyntax;

import org.recommender101.data.DataModel;
import org.recommender101.data.Rating;
import org.recommender101.eval.interfaces.RecommendationlistEvaluator;
import org.recommender101.tools.Utilities101;

import oeg.rs.utils.DamerauLevenshteinDistance;
import oeg.rs.utils.ManhattanDistnce;

/**
 * @author oescamil
 *
 */
public class MEREROEvaluator extends RecommendationlistEvaluator {
	
	protected float errorAccumulator = 0 ; 
	protected int rankingCounter = 0;
	@Override
	public void initialize() {
		super.initialize();
		this.errorAccumulator = 0;
		this.rankingCounter = 0;
	}
	
	
	/* (non-Javadoc)
	 * @see org.recommender101.eval.interfaces.RecommendationlistEvaluator#addRecommendations(java.lang.Integer, java.util.List)
	 */
	@Override
	public void addRecommendations(Integer user, List<Integer> list) {
		DataModel testModel = getTestDataModel();
		DataModel trainingModel = getTrainingDataModel();
		Set<Rating> userRatings = testModel.getRatingsOfUser(user);
		
		// Get rating for all items we know
		Map<Integer, Float> originalRatings = new HashMap<Integer, Float>();
		for(Rating r : userRatings){
			float trainingR = trainingModel.getRating(r.user, r.item); 
			if(trainingR == -1) // Just for items which recommender can recommend
				originalRatings.put(r.item, r.rating);
		}
		originalRatings = Utilities101.sortByKeyAscending(originalRatings);
		originalRatings = Utilities101.sortByValueDescending(originalRatings);
		List<Integer> originalRanking = new ArrayList<Integer>();
		
		int topN = Math.min(this.getTopN(),list.size());
		int counter = 0 ;
		for (Integer item : originalRatings.keySet()) {
			originalRanking.add(item);
			counter++;
			if(counter >= topN)
				break;
		}
		topN = Math.min(topN,originalRanking.size());
		
		list = list.subList(0, topN);
		originalRanking = originalRanking.subList(0, topN);
		
		float error = calcDistance(originalRanking, list);
		
		this.errorAccumulator += error;
		this.rankingCounter++;
		System.gc();
	}

	
	
	
	/* (non-Javadoc)
	 * @see org.recommender101.eval.interfaces.RecommendationlistEvaluator#getEvaluationResult()
	 */
	@Override
	public float getEvaluationResult() {
		if(this.rankingCounter <= 0)
			return 0;
		
		return (float) Math.sqrt(this.errorAccumulator/this.rankingCounter);
	}

	
	
	/**
	 * Used for evaluation aggregation
	 */
	public String toString() {
		return "Ranking Evaluator";
	}
	
	
	
	
	
	/**
	 * Calculates the distance between two rankings (a and b)
	 * @param a Ranked items
	 * @param b Ranked items
	 * @return
	 */
	public static float calcDistance(List<Integer> a, List<Integer> b){
		// For Manhattan distance 
		Map<Integer, Integer> mapA = convertRankingToPositionVector(a);
		Map<Integer, Integer> mapB = convertRankingToPositionVector(b);
		
		
		ManhattanDistnce dlDistance = new ManhattanDistnce();
		Integer[] tmpA = new Integer[0];
		return dlDistance.calc(mapA.values().toArray(tmpA),mapB.values().toArray(tmpA));
	}
	
	
	/**
	 * 
	 * @param ranking
	 * @return
	 */
	public static Map<Integer, Integer> convertRankingToPositionVector(List<Integer> ranking){
		Map<Integer, Integer> res = new HashMap<Integer,Integer>();
		for (int i = 0; i < ranking.size();i++) {
			res.put(ranking.get(i), i+1);
		}
		return res;
	}

	

}
